use std::{
    any::{Any, TypeId},
    collections::HashMap,
};

use anyhow::{Error, anyhow};
use utilities::prelude::{remove_life_time, remove_life_time_mut};

use crate::get_disjoint_mut::GetDisjointMut;

pub trait Resource: Any + Send + Sync {}

impl dyn Resource {
    pub fn downcast_owned<T: Resource + 'static>(mut self: Box<Self>) -> Option<T> {
        self.downcast_mut()
            .map(|raw| unsafe { *Box::from_raw(raw as *mut T) })
    }

    pub fn downcast_ref<T: Resource + 'static>(&self) -> Option<&T> {
        (self as &dyn Any).downcast_ref()
    }

    pub fn downcast_mut<T: Resource + 'static>(&mut self) -> Option<&mut T> {
        (self as &mut dyn Any).downcast_mut()
    }
}

#[derive(Default)]
pub struct Resources {
    map: HashMap<TypeId, Box<dyn Resource>>,
}

impl Resources {
    pub fn insert<T: Resource>(&mut self, value: T) -> Option<T> {
        self.map
            .insert(TypeId::of::<T>(), Box::new(value))
            .map(|any| any.downcast_owned())
            .flatten()
    }

    pub(crate) fn insert_resource(&mut self, type_id: TypeId, resource: Box<dyn Resource>) {
        self.map.insert(type_id, resource);
    }

    pub fn insert_if_not_exists<T: Resource + Default>(&mut self) {
        if !self.contains::<T>() {
            self.insert(T::default());
        }
    }

    pub fn remove<T: Resource>(&mut self) -> Option<T> {
        self.map
            .remove(&TypeId::of::<T>())
            .map(|any| any.downcast_owned())
            .flatten()
    }

    pub fn get<T: Resource>(&self) -> &T {
        self.get_opt::<T>().unwrap()
    }

    pub fn get_by_type_id<T: Resource>(&self, type_id: TypeId) -> &T {
        self.get_opt_by_type_id(type_id).unwrap()
    }

    pub fn get_unchecked<'a, T: Resource>(&self) -> &'a T {
        unsafe { remove_life_time(self.get::<T>()) }
    }

    pub fn get_opt<T: Resource>(&self) -> Option<&T> {
        self.get_opt_by_type_id(TypeId::of::<T>())
    }

    pub fn get_opt_by_type_id<T: Resource>(&self, type_id: TypeId) -> Option<&T> {
        debug_assert_eq!(type_id, TypeId::of::<T>());

        self.map
            .get(&type_id)
            .map(|any| any.downcast_ref())
            .flatten()
    }

    pub fn get_mut_by_type_id<T: Resource>(&mut self, type_id: TypeId) -> &mut T {
        self.get_mut_opt_by_type_id(type_id).unwrap()
    }

    pub fn get_mut_by_type_id_untyped(&mut self, type_id: TypeId) -> &mut dyn Resource {
        self.get_mut_opt_by_type_id_untyped(type_id).unwrap()
    }

    pub fn get_mut_unchecked<'a, T: Resource>(&mut self) -> &'a mut T {
        unsafe { remove_life_time_mut(GetDisjointMut::<'_, &'_ mut T>::get_mut(self).unwrap()) }
    }

    pub fn get_mut_opt<T: Resource>(&mut self) -> Option<&mut T> {
        self.get_mut_opt_by_type_id(TypeId::of::<T>())
    }

    pub fn get_mut_opt_by_type_id<T: Resource>(&mut self, type_id: TypeId) -> Option<&mut T> {
        debug_assert_eq!(type_id, TypeId::of::<T>());

        self.map
            .get_mut(&type_id)
            .map(|any| any.downcast_mut())
            .flatten()
    }

    pub fn get_mut_opt_by_type_id_untyped(&mut self, type_id: TypeId) -> Option<&mut dyn Resource> {
        self.map.get_mut(&type_id).map(|any| any.as_mut())
    }

    pub fn contains<T: Resource>(&self) -> bool {
        self.map.contains_key(&TypeId::of::<T>())
    }
}

impl<'a, T> GetDisjointMut<'a, &'a mut T> for Resources
where
    T: Resource,
{
    type Error = Error;

    fn get_mut(&'a mut self) -> std::result::Result<&'a mut T, Self::Error> {
        self.get_mut_opt()
            .ok_or_else(|| anyhow!("failed downcasting {}", stringify!(T)))
    }
}

macro_rules! impl_get_disjoint_mut {
    ($struct:ident<$($t:ident$(,)?)+>{$error:ident}) => {
        impl<'a, $($t,)+> GetDisjointMut<'a, ($(&'a mut $t,)+)> for $struct
        where
            $(
                $t: Resource + 'static,
            )+
        {
            type Error = $error;

            fn get_mut(&'a mut self) -> std::result::Result<($(&'a mut $t,)+), Self::Error> {
                let mut types: std::collections::VecDeque<_>
                    = self.map.get_disjoint_mut([$(&TypeId::of::<$t>(),)+]).into_iter().collect();

                Ok(($(
                    types
                        .pop_front()
                        .flatten()
                        .map(|any| any.downcast_mut())
                        .flatten()
                        .ok_or_else(|| anyhow!("failed downcasting {}", stringify!($t)))?,
                )+))
            }
        }
    };
}

impl_get_disjoint_mut!(Resources < T > { Error });
impl_get_disjoint_mut!(Resources < T, U > { Error });
impl_get_disjoint_mut!(Resources < T, U, V > { Error });
impl_get_disjoint_mut!(Resources < T, U, V, W > { Error });
impl_get_disjoint_mut!(Resources < T, U, V, W, X > { Error });
impl_get_disjoint_mut!(Resources < T, U, V, W, X, Y > { Error });
impl_get_disjoint_mut!(Resources < T, U, V, W, X, Y, Z > { Error });
impl_get_disjoint_mut!(Resources < T, U, V, W, X, Y, Z, A > { Error });
impl_get_disjoint_mut!(Resources < T, U, V, W, X, Y, Z, A, B > { Error });
impl_get_disjoint_mut!(Resources < T, U, V, W, X, Y, Z, A, B, C > { Error });
